package com.atlassian.crucible.plugin.export;

import java.awt.Color;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.ResourceBundle;
import java.util.Set;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import com.atlassian.crucible.plugin.export.exception.ExportException;
import com.atlassian.crucible.plugin.export.exception.PdfCreationException;
import com.atlassian.crucible.plugin.export.exception.ReviewDataRetrievalException;
import com.atlassian.crucible.spi.PermId;
import com.atlassian.crucible.spi.data.FisheyeReviewItemData;
import com.atlassian.crucible.spi.data.GeneralCommentData;
import com.atlassian.crucible.spi.data.ReviewData;
import com.atlassian.crucible.spi.data.ReviewItemData;
import com.atlassian.crucible.spi.data.VersionedLineCommentData;
import com.atlassian.crucible.spi.data.ReviewData.State;
import com.lowagie.text.Chunk;
import com.lowagie.text.Document;
import com.lowagie.text.DocumentException;
import com.lowagie.text.Font;
import com.lowagie.text.Image;
import com.lowagie.text.Paragraph;
import com.lowagie.text.Phrase;
import com.lowagie.text.Rectangle;
import com.lowagie.text.pdf.BaseFont;
import com.lowagie.text.pdf.PdfContentByte;
import com.lowagie.text.pdf.PdfPTable;
import com.lowagie.text.pdf.PdfPageEventHelper;
import com.lowagie.text.pdf.PdfTemplate;
import com.lowagie.text.pdf.PdfWriter;

/**
 * Handles exporting a Crucible review to PDF, using the
 * iText(http://itextdocs.lowagie.com/) library.
 * 
 * @author Ross Rowe
 * 
 */
public class ReviewExporter extends PdfPageEventHelper {

	private static final String HEADER_LOGO = "/download/resources/com.atlassian.crucible.plugin.export/combined_header_logo.png";
	private static final String KEY = "crucible.export.";
	

	private static final Logger logger = Logger.getLogger(ReviewExporter.class);

	private static BaseFont HELVETICA;
	private static Font HELVETICA_14;
	private static Font HELVETICA_18;
	private static Font HELVETICA_BOLD;
	private static Font HELVETICA_6;

	private PdfTemplate total;
	private CrucibleClient client;
	private String baseUrl;
	private final Date generatedDate = new Date();

	private ResourceBundle messages;
	private static final String FILE_COMMENTS = KEY + "fileComments";
	private static final String LINES = KEY + "lines";
	private static final String DEFECT = KEY + "defect";
	private static final String YES = KEY + "yes";
	private static final String NO = KEY + "no";
	private static final String FILE = KEY + "file";
	private static final String REPORT_GENERATED = KEY + "reportGenerated";
	private static final String REVIEW_DETAILS = KEY + "reviewDetails";
	private static final String OBJECTIVES = KEY + "objectives";
	private static final String AUTHOR = KEY + "author";
	private static final String MODERATOR = KEY + "moderator";
	private static final String STATE = KEY + "state";
	private static final String PAGE = KEY + "page";
	private static final String GENERAL_COMMENTS = KEY + "generalComments";
	private static final String TITLE = KEY + "title";
	private static final String USER = KEY + "user";
	private static final String COMMENT = KEY + "comment";
	private static final String PASSWORD_BLANK = KEY + "error.blank.password";
	private static final String USER_NAME_BLANK = KEY + "error.blank.user";
	private static final String REVIEW_ID_BLANK = KEY + "error.blank.reviewId";

	/**
	 * Mapping of {@link VersionedLineCommentData} and
	 * {@link FisheyeReviewItemData} instances.
	 * 
	 * @author Ross Rowe
	 * 
	 */
	public class ReviewMap {

		private VersionedLineCommentData comment;
		private FisheyeReviewItemData reviewItemData;

		public ReviewMap(VersionedLineCommentData comment,
				FisheyeReviewItemData reviewItemData) {
			this.comment = comment;
			this.reviewItemData = reviewItemData;
		}

		public VersionedLineCommentData getComment() {
			return comment;
		}

		public FisheyeReviewItemData getReviewItemData() {
			return reviewItemData;
		}

	}

	static {
		try {
			HELVETICA = BaseFont.createFont(BaseFont.HELVETICA,
					BaseFont.WINANSI, BaseFont.NOT_EMBEDDED);
			HELVETICA_BOLD = new Font(HELVETICA, Font.DEFAULTSIZE, Font.BOLD);
			HELVETICA_6 = new Font(HELVETICA, 6);
			HELVETICA_14 = new Font(HELVETICA, 14);
			HELVETICA_18 = new Font(HELVETICA, 18);
		} catch (DocumentException e) {
			logger.error(e);
		} catch (IOException e) {
			logger.error(e);
		}
	}

	/**
	 * Constructs a ReviewExporter instance.
	 * 
	 * @param client
	 */
	public ReviewExporter(CrucibleClient client, Locale locale) {
		this.client = client;
		this.messages = ResourceBundle.getBundle("ReviewExporter", locale);
	}

	/**
	 * Returns the {@link CrucibleClient} instance.
	 * 
	 * @return
	 */
	public CrucibleClient getClient() {
		return client;
	}

	/**
	 * Performs input validation checks, retrieves the {@link ReviewData},
	 * {@link GeneralCommentData} and {@link VersionedLineCommentData} for the
	 * specific reviewId, and invokes the PDF processing.
	 * 
	 * @param userName
	 * @param password
	 * @param reviewId
	 * @return
	 * @throws ExportException
	 */
	public ByteArrayOutputStream createPdfExport(String userName,
			String password, String reviewId) throws ExportException {
		checkInputData(userName, password, reviewId);
		ReviewData reviewData = getReviewData(userName, password, reviewId);
		List<GeneralCommentData> generalComments = getGeneralComments(reviewId);
		Map<String, List<ReviewMap>> reviewMap = getCommentMap(reviewId);
		return createPDF(reviewData, generalComments, reviewMap);
	}

	/**
	 * Iterates through the list of versioned comments for a reviewId, and
	 * builds a map (keyed on the file being commented upon) containing a list
	 * of {@link ReviewMap} instances.
	 * 
	 * @param reviewId
	 * @return
	 */
	private Map<String, List<ReviewMap>> getCommentMap(String reviewId) {
		Map<String, List<ReviewMap>> map = new HashMap<String, List<ReviewMap>>();

		List<VersionedLineCommentData> versionedComments = getVersionedComments(reviewId);
		for (VersionedLineCommentData comment : versionedComments) {
			PermId<ReviewItemData> reviewItemId = comment.getReviewItemId();
			if (reviewItemId != null) {
				FisheyeReviewItemData reviewItemData = client
						.getReviewItemData(reviewId, reviewItemId.getId());
				List<ReviewMap> comments = map.get(reviewItemData.getToPath());
				if (comments == null) {
					comments = new ArrayList<ReviewMap>();
				}
				comments.add(new ReviewMap(comment, reviewItemData));
				map.put(reviewItemData.getToPath(), comments);
			}
		}
		return map;
	}

	/**
	 * Returns the list of {@link GeneralCommentData} instances for the
	 * reviewId.
	 * 
	 * @param reviewId
	 * @return
	 */
	private List<GeneralCommentData> getGeneralComments(String reviewId) {
		return client.getGeneralComments(reviewId);
	}

	/**
	 * Returns a list of {@link VersionedLineCommentData} instances for the
	 * reviewId.
	 * 
	 * @param reviewId
	 * @return
	 */
	private List<VersionedLineCommentData> getVersionedComments(String reviewId) {
		return client.getVersionedComments(reviewId);
	}

	/**
	 * Validates that the input variables are not null or blank.
	 * 
	 * @param userName
	 * @param password
	 * @param reviewId
	 * @throws ExportException
	 */
	private void checkInputData(String userName, String password,
			String reviewId) throws ExportException {
		if (!StringUtils.isNotBlank(userName))
			throw new ReviewDataRetrievalException(getText(USER_NAME_BLANK));
		if (!StringUtils.isNotBlank(password))
			throw new ReviewDataRetrievalException(getText(PASSWORD_BLANK));
		if (!StringUtils.isNotBlank(reviewId))
			throw new ReviewDataRetrievalException(getText(REVIEW_ID_BLANK));
	}

	/**
	 * Retrieves the {@link ReviewData} for the reviewId.
	 * 
	 * @param userName
	 * @param password
	 * @param reviewId
	 * @return
	 */
	private ReviewData getReviewData(String userName, String password,
			String reviewId) {
		client.login(userName, password);
		return client.getReview(reviewId);
	}

	/**
	 * Creates a new PdfWriter, invokes the processing of the Pdf and returns
	 * the outputStream.
	 * 
	 * @param reviewData
	 * @param generalComments
	 * @param reviewMap
	 * @return
	 * @throws ExportException
	 */
	private ByteArrayOutputStream createPDF(ReviewData reviewData,
			List<GeneralCommentData> generalComments,
			Map<String, List<ReviewMap>> reviewMap) throws ExportException {
		if (reviewData == null)
			throw new ReviewDataRetrievalException("Review Data is null");
		Document document = new Document();
		try {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			PdfWriter writer = PdfWriter.getInstance(document, baos);
			writer.setPageEvent(this);
			document.open();
			writeSummaryInfo(reviewData, document);
			writeGeneralComments(generalComments, document);
			writeVersionedComments(reviewMap, document, reviewData.getPermaId()
					.getId());
			document.close();
			return baos;
		} catch (DocumentException e) {
			throw new PdfCreationException(e);
		}
	}

	/**
	 * Iterates over the generalComments list, and adds the data to the PDF
	 * output.
	 * 
	 * @param generalComments
	 * @param document
	 * @throws DocumentException
	 */
	private void writeGeneralComments(List<GeneralCommentData> generalComments,
			Document document) throws DocumentException {
		Chunk spacerChunk = new Chunk("   ", HELVETICA_18);
		document.add(new Paragraph(spacerChunk));
		Chunk headerChunk = new Chunk(getText(GENERAL_COMMENTS), HELVETICA_18);
		document.add(new Paragraph(headerChunk));
		document.add(new Paragraph(new Chunk("   ", HELVETICA_14)));

		float[] widths = { 1f, 4f };
		PdfPTable table = new PdfPTable(widths);
		table.setTotalWidth(420f);
		table.setWidthPercentage(100);

		addHeaderCell(table, getText(USER));
		addHeaderCell(table, getText(COMMENT));
		for (GeneralCommentData comment : generalComments) {
			table.addCell(comment.getUser());
			table.addCell(comment.getMessage());
		}
		document.add(table);
	}

	/**
	 * Iterates over the reviewMap key set, and adds a table that includes the
	 * comment data for each file to the PDF output.
	 * 
	 * @param reviewMap
	 * @param document
	 * @param reviewId
	 * @throws DocumentException
	 */
	private void writeVersionedComments(Map<String, List<ReviewMap>> reviewMap,
			Document document, String reviewId) throws DocumentException {
		Chunk spacerChunk = new Chunk("   ", HELVETICA_18);
		document.add(new Paragraph(spacerChunk));
		Chunk headerChunk = new Chunk(getText(FILE_COMMENTS), HELVETICA_18);
		document.add(new Paragraph(headerChunk));

		// table.setLockedWidth(true);
		Set<String> keys = reviewMap.keySet();
		for (String fileName : keys) {
			document.add(new Paragraph(new Chunk(getText(FILE, new Object[]{fileName}),
					HELVETICA_14)));

			float[] widths = { 1f, 1f, 4f, 1f };
			PdfPTable table = new PdfPTable(widths);
			table.setTotalWidth(420f);
			table.setWidthPercentage(100);

			addHeaderCell(table, getText(LINES));
			addHeaderCell(table, getText(USER));
			addHeaderCell(table, getText(COMMENT));
			addHeaderCell(table, getText(DEFECT));

			List<ReviewMap> reviewMapItemList = reviewMap.get(fileName);
			for (ReviewMap reviewMapItem : reviewMapItemList) {
				VersionedLineCommentData comment = reviewMapItem.getComment();
				if (!comment.isDraft()) {
					table.addCell(comment.getToLineRange());
					table.addCell(getUserName(comment.getUser()));
					table.addCell(comment.getMessage());
					table.addCell(comment.isDefectRaised() ? getText(YES)
							: getText(NO));
				}
			}
			document.add(new Paragraph(new Chunk("  ", HELVETICA_6)));
			document.add(table);
		}
	}

	/**
	 * Performs a lookup of the user's display name.
	 * 
	 * @param userId
	 * @return
	 */
	private String getUserName(String userId) {
		return client.getUserName(userId);
	}

	/**
	 * Writes the review summary information to the PDF output.
	 * 
	 * @param reviewData
	 * @param document
	 * @throws DocumentException
	 */
	private void writeSummaryInfo(ReviewData reviewData, Document document)
			throws DocumentException {
		Chunk spacerChunk = new Chunk("   ", HELVETICA_18);
		document.add(new Paragraph(spacerChunk));
		document.add(new Paragraph(spacerChunk));
		Chunk headerChunk = new Chunk(getText(REVIEW_DETAILS), HELVETICA_18);
		document.add(new Paragraph(headerChunk));
		document.add(new Paragraph(spacerChunk));

		float[] widths = { 1f, 4f };
		PdfPTable table = new PdfPTable(widths);
		table.setTotalWidth(420f);
		table.setWidthPercentage(100);
		addHeaderCell(table, getText(TITLE));
		table.addCell(reviewData.getName());
		addHeaderCell(table, getText(OBJECTIVES));
		table.addCell(reviewData.getDescription());
		addHeaderCell(table, getText(AUTHOR));
		table.addCell(reviewData.getAuthor());
		addHeaderCell(table, getText(MODERATOR));
		table.addCell(reviewData.getModerator());

		State state = reviewData.getState();
		if (state != null) {
			addHeaderCell(table, getText(STATE));
			table.addCell(state.name());
		}
		document.add(table);
	}

	/**
	 * Adds a cell to the table in a bold font.
	 * 
	 * @param table
	 * @param text
	 */
	private void addHeaderCell(PdfPTable table, String text) {
		Chunk chunk = new Chunk(text, HELVETICA_BOLD);
		table.addCell(new Phrase(chunk));
	}

	/**
	 * Adds the Crucible image to the top of the page
	 */
	public void onOpenDocument(PdfWriter writer, Document document) {
		total = writer.getDirectContent().createTemplate(100, 100);
		total.setBoundingBox(new Rectangle(-20, -20, 100, 100));
		try {
			Paragraph p = new Paragraph();
			Image image = Image
					.getInstance(baseUrl
							+ HEADER_LOGO);
			image.setBorder(Image.BOX);
			image.setBorderColor(new Color(0xFF, 0x00, 0x00));
			image.setBorderWidth(5);
			p.add(new Chunk(image, 0, -15));
			document.add(p);
		} catch (DocumentException e) {
			logger.error(e);
		} catch (IOException e) {
			logger.error(e);
		}
	}

	/**
	 * Displays the page number and page generation information.
	 */
	public void onEndPage(PdfWriter writer, Document document) {
		showPageNumbering(writer, document);
		showGeneratedInformation(writer, document);
	}

	/**
	 * Displays the page generation information on the bottom left of the page.
	 * 
	 * @param writer
	 * @param document
	 */
	private void showGeneratedInformation(PdfWriter writer, Document document) {
		PdfContentByte contentByte = writer.getDirectContent();
		contentByte.saveState();
		String text = getText(REPORT_GENERATED, new Object[]{generatedDate});
		float textBase = document.bottom() - 20;
		float textSize = HELVETICA.getWidthPoint(text, 12);
		contentByte.beginText();
		contentByte.setFontAndSize(HELVETICA, 12);
		float adjust = HELVETICA.getWidthPoint("0", 12);
		contentByte.setTextMatrix(document.right() - textSize - adjust,
				textBase);
		contentByte.showText(text);
		contentByte.endText();
		contentByte.addTemplate(total, document.right() - adjust, textBase);
		contentByte.restoreState();
	}

	/**
	 * Displays the page numbering on the bottom right of the page.
	 * 
	 * @param writer
	 * @param document
	 */
	private void showPageNumbering(PdfWriter writer, Document document) {
		PdfContentByte contentByte = writer.getDirectContent();
		contentByte.saveState();
		String text = getText(PAGE,new Object[]{writer.getPageNumber()});
		float textBase = document.bottom() - 20;
		float textSize = HELVETICA.getWidthPoint(text, 12);
		contentByte.beginText();
		contentByte.setFontAndSize(HELVETICA, 12);
		contentByte.setTextMatrix(document.left(), textBase);
		contentByte.showText(text);
		contentByte.endText();
		contentByte.addTemplate(total, document.left() + textSize, textBase);
		contentByte.restoreState();
	}

	/**
	 * Finishes the calculation of the page numbering.
	 */
	public void onCloseDocument(PdfWriter writer, Document document) {
		total.beginText();
		total.setFontAndSize(HELVETICA, 12);
		total.setTextMatrix(0, 0);
		total.showText(String.valueOf(writer.getPageNumber() - 1));
		total.endText();
	}

	/**
	 * Sets the baseUrl variable.
	 * 
	 * @param baseUrl
	 */
	public void setBaseUrl(String baseUrl) {
		this.baseUrl = baseUrl;
	}

	private String getText(String key) {
		return messages.getString(key);
	}

	private String getText(String key, Object[] args) {
		MessageFormat formatter = new MessageFormat("");
		formatter.setLocale(messages.getLocale());
		formatter.applyPattern(messages.getString(key));
		return formatter.format(args);

	}

}
